"""
PCDownloader class for downloading data from the Planetary Computer STAC API.
"""

from ..STACDownloader import STACDownloader
from typing import Optional, Dict


class PCDownloader(STACDownloader):
    def __init__(self, aoi, datetime, query: Optional[Dict[str, Dict]] = None):
        try:
            from planetary_computer import sign_inplace
        except ImportError:
            raise ImportError(
                "The planetary_computer package is required. Please install it with 'pip install planetary-computer' and try again."
            )

        self.url = "https://planetarycomputer.microsoft.com/api/stac/v1"
        self.modifier = sign_inplace
        super().__init__(aoi, datetime, query)



"""
STACDownloader class to download and load satellite data from a STAC API.
"""

import geopandas as gpd

from datetime import datetime
from typing import List, Union, Optional

from .decorators import with_rioxarray


class STACDownloader:
    def __init__(
        self,
        aoi: Union[str, gpd.GeoDataFrame],
        datetime: List[Union[str, datetime]],
        query: Optional[dict] = None,
    ):
        try:
            from odc.stac import configure_rio, load

            self.load = load
        except ImportError:
            raise ImportError(
                "The odc.stac package is required. Please install it with 'pip install odc-stac' and try again."
            )
        try:
            import pystac_client
        except ImportError:
            raise ImportError(
                "The pystac_client package is required. Please install it with 'pip install pystac-client' and try again."
            )
        configure_rio(cloud_defaults=True, aws={"aws_unsigned": True})

        self.aoi = aoi
        self.datetime = datetime
        self.query = query
        self.catalog = pystac_client.Client.open(self.url, modifier=self.modifier)

    def search_stac(self):
        items = self.catalog.search(
            bbox=self.aoi,
            datetime=self.datetime,
            collections=self.collection,
            query=self.query,
        ).item_collection()

        if len(items) == 0:
            print(
                f"No images found for {self.datetime} in {self.collection} collection"
            )
            return None

        return items

    def load_stac(
        self,
        groupby: Optional[str] = "solar_day",
        chunks: Optional[dict] = {"time": 5, "x": 512, "y": 512},
    ):
        items = self.search_stac()
        if not items:
            return None
        data = self.load(
            items,
            chunks=chunks,
            crs=self.crs,
            bands=self.bands,
            resolution=self.resolution,
            groupby=groupby,
            bbox=self.aoi,
        )

        return data

    @staticmethod
    @with_rioxarray
    def clip_data(data, gdf):
        return data.rio.clip(gdf.geometry.values, gdf.crs, drop=False)



def with_rio(func):
    def wrapper(self, *args, **kwargs):
        try:
            import rasterio as rio

            return func(self, rio, *args, **kwargs)
        except ImportError:
            raise ImportError(
                "The rasterio package is required for this function. Please install it with 'pip install rasterio' and try again."
            )

    return wrapper


def with_xarray(func):
    def wrapper(self, *args, **kwargs):
        try:
            import xarray as xr

            return func(self, xr, *args, **kwargs)
        except ImportError:
            raise ImportError(
                "The xarray package is required for this function. Please install it with 'pip install xarray' and try again."
            )

    return wrapper


def with_geopandas(func):
    def wrapper(self, *args, **kwargs):
        try:
            import geopandas as gpd

            return func(self, gpd, *args, **kwargs)
        except ImportError:
            raise ImportError(
                "The geopandas package is required for this function. Please install it with 'pip install geopandas' and try again."
            )

    return wrapper


def with_rioxarray(func):
    def wrapper(self, *args, **kwargs):
        try:
            import rioxarray as rxr

            return func(self, rxr, *args, **kwargs)
        except ImportError:
            raise ImportError(
                "The rioxarray package is required for this function. Please install it with 'pip install rioxarray' and try again."
            )

    return wrapper



import os
from .BaseStorage import BaseStorage
from .S3Storage import S3Storage
from .LocalStorage import LocalStorage


class Storage(BaseStorage):
    def __init__(self):
        super().__init__()
        self.storages_names = os.environ.get("SPAI_STORAGE_NAMES", None)
        self.storages = {}
        if self.storages_names:
            self.storages_names = self.storages_names.split(",")
            self.initialize_storage()

    def __getitem__(self, name):
        try:
            return self.storages[name]
        except KeyError:
            raise KeyError(f"Storage '{name}' not found")

    def initialize_storage(self):
        for storage_name in self.storages_names:
            envs = {
                key: value
                for key, value in os.environ.items()
                if key.startswith(f"SPAI_STORAGE_{storage_name.upper()}")
            }
            envs = {key.split("_")[-1].lower(): value for key, value in envs.items()}
            if storage_name.split("_")[0] == "local":
                self.initialize_local(storage_name, envs)
            elif storage_name.split("_")[0] == "s3":
                self.initialize_s3(storage_name, envs)

    def initialize_local(self, name, envs):
        self.storages[f"{name.split('_')[1]}"] = LocalStorage(**envs)

    def initialize_s3(self, name, envs):
        self.storages[f"{name.split('_')[1]}"] = S3Storage(**envs)



import os
import geopandas as gpd
import geojson
from warnings import warn
from shapely.geometry import box
from shapely.geometry.polygon import Polygon
import requests
from datetime import datetime, timedelta
from dateutil.parser import isoparse
from typing import Optional, Tuple


def validate_coords(coords):
    # Coordinates of points of a polygon: [[lon1,lat1],[lon2,lat2], ... ,[lonN,latN]]
    for coord in coords:
        # check that each coord is a list of 2 coordinates
        if len(coord) != 2:
            raise Exception("each coord must be a list of 2 coordinates")
        # check that each coordinate is a float
        if not isinstance(coord[0], float) or not isinstance(coord[1], float):
            raise Exception("each coordinate must be a float")
        lon = coord[0]
        lat = coord[1]

        # check lat and long ranges
        if lon < -180 or lon > 180 or lat < -90 or lat > 90:
            raise Exception("each coordinate must be a valid lat/long")

    return coords


def validate_bounds(bounds):
    # check that bounds is a list
    if len(bounds) != 4 or not isinstance(bounds, tuple):
        raise Exception(
            "bounds must be a tuple of 4 points: (minlon, minlat, maxlon, maxlat))"
        )
    # check that each bound is a float
    for bound in bounds:
        if not isinstance(bound, float):
            raise Exception("each bound must be a float")
    # check lat and long ranges
    minlon = bounds[0]
    minlat = bounds[1]
    maxlon = bounds[2]
    maxlat = bounds[3]

    if minlat < -90 or minlat > 90 or maxlat < -90 or maxlat > 90:
        raise Exception("each latitude must be a valid lat/long")
    if minlon < -180 or minlon > 180 or maxlon < -180 or maxlon > 180:
        raise Exception("each longitude must be a valid lat/long")
    # check that minlat < maxlat and minlng < maxlng
    if minlon > maxlon or minlat > maxlat:
        raise Exception(
            "minlat must be less than maxlat and minlng must be less than maxlng"
        )
    return bounds


def create_aoi_geodataframe(obj, crs):
    # obj can be a :
    # GeoDataFrame,
    # GeoJSON file,
    # Path to a GeoJSON file,
    # Polygon of Shapely (Bounding Box),
    # List of Coords of a (Bounding Box),
    # Location name (string)

    # return a GeoDataFrame of the Bounding Box of the obj or a list of gdfs if obj is a location name and there are more than one relation
    gdf = gpd.GeoDataFrame()

    if (
        isinstance(obj, str)
        and not os.path.isfile(obj)
        and not obj.endswith(".geojson")
    ):
        # case 1: obj is a location name (e.g. "Madrid")
        gdf = get_box_or_gdfs_by_place_name(
            obj
        )  # return GeoDataFrame of the Bounding Box of the location

    elif isinstance(obj, list):
        # case 2: obj is a list of coords of a Polygon: [[lat1,long1],[lat2,long2],...,[latN,longN]]
        coords = validate_coords(obj)
        poly = Polygon(coords)
        bounds = poly.bounds
        obj = box(*bounds)  # return Box of Shapely

    if isinstance(obj, tuple):
        # case 3: obj is a tuple of bounds: (minlat, minlng, maxlat, maxlng)
        bounds = validate_bounds(obj)
        obj = box(*bounds)  # return Box of Shapely

    if isinstance(obj, Polygon):
        # case 4: obj is a Box of Shapely: Polygon[[lat1,long1],[lat2,long2],[lat3,long3],[lat4,long4]]
        bounds = obj.bounds
        validate_bounds(bounds)
        gdf = gpd.GeoDataFrame(geometry=[obj], crs=4326)

    if isinstance(obj, gpd.GeoDataFrame):
        # case 5: obj is a GeoDataFrame
        gdf = obj

    if (
        isinstance(obj, dict)
        and obj.get("type") == "FeatureCollection"
        and "features" in obj
    ):
        # case 6: obj is a GeoJSON file
        if obj["type"] == "Polygon":
            for coords in obj["coordinates"][0]:
                validate_coords(coords)
        elif obj["type"] == "MultiPolygon":
            for _coords in obj["coordinates"][0]:
                for coords in _coords:
                    validate_coords(coords)
        gdf = gpd.GeoDataFrame.from_features(obj, crs=4326)

    if isinstance(obj, str) and os.path.isfile(obj) and obj.endswith(".geojson"):
        # case 7: obj is a path to GeoJSON file
        geojson_file = geojson.load(open(obj))
        if geojson_file["type"] == "Polygon":
            for coords in geojson_file["coordinates"][0]:
                validate_coords(coords)
        elif geojson_file["type"] == "MultiPolygon":
            for _coords in geojson_file["coordinates"][0]:
                for coords in _coords:
                    validate_coords(coords)
        gdf = gpd.GeoDataFrame.from_features(geojson_file, crs=4326)

    if not gdf.crs:
        warn("GeoDataFrame has no crs, assuming EPSG:4326")
        gdf = gdf.set_crs(epsg=4326)
    if gdf.crs != crs:
        gdf = gdf.to_crs(crs)

    if gdf.empty:
        raise Exception(f"Location {obj} not supported")

    if is_valid_geodataframe(gdf):
        return gdf


def get_bb_by_city_name(city_name):
    base_url = "https://nominatim.openstreetmap.org"
    format_out = "json"
    limit = 10

    # Construct the API request URL
    url = f"{base_url}/search?city={city_name}&format={format_out}&limit={limit}"

    # Send the API request
    response = requests.get(url).json()

    results = []
    if len(response) == 0:
        raise Exception("No results found")
    for result in response:
        if "boundingbox" in result:
            bounding_box = result["boundingbox"]
            if len(bounding_box) == 4:
                min_lon, max_lon, min_lat, max_lat = map(float, bounding_box)
                results.append(
                    {
                        "name": f"{result['display_name']}",
                        "bbox": box(min_lat, min_lon, max_lat, max_lon),
                    }
                )

    return results


def get_box_or_gdfs_by_place_name(place_name):
    if not isinstance(place_name, str):
        raise Exception("place_name must be a string")
    results = get_bb_by_city_name(place_name)

    if len(results) == 0:
        return None
    else:
        # Always return the first result as gdf
        first_result = results[0]
        gdf = gpd.GeoDataFrame({"name": [place_name]})
        gdf.set_geometry([first_result["bbox"]], inplace=True)
        gdf.set_crs(epsg=4326, inplace=True)

        return gdf


def is_valid_geodataframe(gdf: gpd.GeoDataFrame) -> bool:
    if len(gdf) != 1:
        raise ValueError("AoI must be a single polygon or area, not multiple polygons")

    if not gdf.geometry.is_valid.all():
        raise ValueError("The given geometry is not valid")

    # Convert to crs 3857 and check area in km2 is less than 10. If not, return False
    gdf_utm = gdf.to_crs(3857)
    area = gdf_utm.area[0] / 10**6
    if area / 10**6 > 20:  # 20 km2
        raise ValueError("Area of AOI is greater than 20 km2")

    return True


def is_valid_datetime_param(param):
    if isinstance(param, datetime):
        return True
    elif isinstance(param, str):
        try:
            isoparse(param)
            return True
        except ValueError:
            try:
                datetime.strptime(param, "%Y-%m-%d")
                return True
            except ValueError:
                try:
                    datetime.strptime(param, "%Y-%m")
                    return True
                except ValueError:
                    try:
                        datetime.strptime(param, "%Y")
                        return True
                    except ValueError:
                        return False
    elif isinstance(param, (list, tuple)) and len(param) == 2:
        if isinstance(param[0], (datetime, str)) and isinstance(
            param[1], (datetime, str)
        ):
            return True
    elif isinstance(param, str) and "/" in param:
        parts = param.split("/")
        if len(parts) == 2:
            for part in parts:
                if part.strip() == "..":
                    return True
                else:
                    try:
                        isoparse(part.strip())
                    except ValueError:
                        return False
            return True
    return False


def get_last_month(starting_date: Optional[str] = None) -> Tuple[str, str]:
    now = datetime.now() if not starting_date else isoparse(starting_date)
    last_months = now - timedelta(days=30)
    time_interval = (last_months.strftime("%Y-%m-%d"), now.strftime("%Y-%m-%d"))

    if is_valid_datetime_param(time_interval):
        return time_interval


def add_item_extra_properties(item, item_dict, query_properties, query_parameters):
    for parameter in query_parameters.keys():
        for prop in query_properties:
            if parameter in prop:
                item_dict[prop] = item.properties[prop]

    return item_dict



from ..repos import APIRepo


def retrieve_credentials(user):
    repo = APIRepo()
    data, error = repo.retrieve_user(user)
    if error:
        raise Exception(error)
    access_key = data.get("access_key")
    secret_key = data.get("secret_key")
    if not access_key or not secret_key:
        raise Exception("No credentials found.")
    return {"AWS_ACCESS_KEY_ID": access_key, "AWS_SECRET_ACCESS_KEY": secret_key}



import os
import schedule
import time
from multiprocessing import Process
from pathlib import Path
import json

from ..config import load_and_validate_config
from ..storage import create_or_retrieve_s3_bucket


def run_item(item, command):
    if item.command:
        os.system(item.command)
    else:
        os.system(command)


def run_script(args):
    script, dir, typer = args
    typer.echo(f"Running script '{script.name}'...")
    os.chdir(dir / f"scripts/{script.name}")  # change to script dir
    run_item(script, f"python main.py")


def run_notebook(notebook, dir, typer):
    typer.echo(f"Running notebook '{notebook.name}'...")
    os.chdir(dir / "notebooks" / notebook.name)
    os.system(f"jupyter notebook")


def run_api(api, dir, port, host):
    os.chdir(dir / f"apis/{api.name}")
    os.system(f"python main.py --port {port} --host {host}")


def run_ui(ui, dir, api_urls):
    os.chdir(dir / f"uis/{ui.name}")
    cmd = ""
    for name, value in ui.env.items():
        if value in api_urls:
            cmd += f"export {name}={api_urls[value]}; "
    cmd += ui.command
    os.system(cmd)


def run_schdule():
    while True:
        schedule.run_pending()
        time.sleep(1)


def run_local(dir, variables, typer):
    config = load_and_validate_config(dir, typer)
    dir = Path(dir).resolve()  # required because of the chdir
    typer.echo(f"Deploying locally...")
    # save variables
    if variables:
        try:
            # load current variables
            with open(dir / "spai.vars.json", "r") as f:
                current_variables = json.load(f)
        except FileNotFoundError:
            current_variables = {}
        # merge variables
        variables = {**current_variables, **variables}
        with open(dir / "spai.vars.json", "w") as f:
            json.dump(variables, f)
    keep_alive = False
    processes = []
    if config.storage:
        typer.echo(f"Setting up storage...")
        for storage in config.storage:
            if storage.type == "s3" and not storage.credentials:
                # raise ValueError("S3 storage requires credentials")
                typer.echo(
                    f"S3 storage credentials not provided, an S3 bucket will be created if it does not already exist."
                )
                # if not typer.confirm("Do you want to continue?"):
                #     raise typer.Abort()
                storage.credentials = create_or_retrieve_s3_bucket(
                    config.project, storage.name
                )
        # set env variables
        prefix = "SPAI_STORAGE"
        names = []
        for storage in config.storage:
            names.append(storage.type + "_" + storage.name)
            name = storage.name.upper()
            if storage.type == "local":
                os.environ[f"{prefix}_LOCAL_{name}_PATH"] = (
                    str(dir) + "/" + storage.path
                )
            elif storage.type == "s3":
                os.environ[f"{prefix}_S3_{name}_URL"] = storage.credentials.url
                os.environ[f"{prefix}_S3_{name}_ACCESS"] = (
                    storage.credentials.access_key
                )
                os.environ[f"{prefix}_S3_{name}_SECRET"] = (
                    storage.credentials.secret_key
                )
                os.environ[f"{prefix}_S3_{name}_REGION"] = storage.credentials.region
                os.environ[f"{prefix}_S3_{name}_BUCKET"] = storage.credentials.bucket
        os.environ[f"{prefix}_NAMES"] = ",".join(names)
    if config.scripts:
        typer.echo(f"Deploying scripts...")
        for script in config.scripts:
            # the order affects in the execution of the scheduled tasks
            if script.run_on_start:
                run_script((script, dir, typer))
            if script.run_every:
                keep_alive = True
                schedule.every(script.run_every).minutes.do(
                    run_script, (script, dir, typer)
                )  # if a task goes after another than takes more time than the scheduled time, it will have to wait
    if config.notebooks:
        typer.echo(f"Deploying notebooks...")
        for notebook in config.notebooks:
            p = Process(target=run_notebook, args=(notebook, dir, typer))
            p.start()
            processes.append(p)
    api_urls = {}
    if config.apis:
        typer.echo(f"Deploying apis...")
        for api in config.apis:
            typer.echo(f"Running api '{api.name}'...")
            p = Process(target=run_api, args=(api, dir, api.port, api.host))
            p.start()
            processes.append(p)
            # save api_urls in case UIs need them
            api_urls[f"api.{api.name}"] = f"{api.host}:{api.port}"
    if config.uis:
        typer.echo(f"Deploying uis...")
        for ui in config.uis:
            typer.echo(f"Running ui '{ui.name}'...")
            p = Process(target=run_ui, args=(ui, dir, api_urls))
            p.start()
            processes.append(p)
    typer.echo(f"Project '{config.project}' deployed.")
    if keep_alive:
        p = Process(target=run_schdule)
        p.start()
        processes.append(p)
    for p in processes:
        p.join()
    return



"""
PCS1GRDDownloader class to download and load Sentinel-1 GRD data from a STAC API of the Planetary Computer.
"""

from .PCDownloader import PCDownloader

from typing import Optional, List


S1_BANDS = [
    "vv",
    "vh",
]


class PCS1GRDDownloader(PCDownloader):
    def __init__(
        self,
        aoi,
        datetime,
        bands: Optional[List[str]] = S1_BANDS,
        resolution: Optional[float] = 0.000045045,
        crs: Optional[str] = "EPSG:4326",
    ):
        # TODO add filters
        # See filters in https://earth-search.aws.element84.com/v1/collections/sentinel-1-grd/items
        query = {}
        super().__init__(aoi, datetime, query)
        self.collection = "sentinel-1-grd"
        self.bands = bands
        self.resolution = resolution
        self.crs = crs



from pydantic import BaseModel, validator
from typing import Optional


class StorageConfig(BaseModel):
    type: str
    name: str


class LocalStorageConfig(StorageConfig):
    path: Optional[str] = None

    @validator("type")
    def type_must_be_local(cls, v):
        if not v == "local":
            raise ValueError("type must be local")
        return v

    @validator("path", pre=True, always=True)
    def user_name_as_path_if_not_path(cls, v, *, values, **kwargs):
        return v or values["name"]


class S3StorageCredentials(BaseModel):
    url: str
    access_key: str
    secret_key: str
    bucket: str
    region: Optional[str] = (
        None  # en local no hace falta, además afecta a la conexión segura o no
    )


class S3StorageConfig(StorageConfig):
    credentials: Optional[S3StorageCredentials] = None

    @validator("type")
    def type_must_be_s3(cls, v):
        if not v == "s3":
            raise ValueError("type must be s3")
        return v



from ..repos import APIRepo
import os


def deploy_template(user, template_name, variables, typer):
    repo = APIRepo()
    data, error = repo.deploy_template(user, template_name, variables)
    if data:
        typer.echo(
            f"Check status at {os.getenv('SPAI_UI_URL', 'https://spai.earthpulse.ai')}/projects/{data['project_id']}"
        )
        return
    raise Exception(f"Error: {error}")



import numpy as np


def control_uint_type(array: np.array) -> np.dtype:
    if not issubclass(array.dtype.type, np.integer):
        # If the array is not an integer, return the same type
        return array.dtype

    min = np.min(array)
    max = np.max(array)

    if min >= 0 and max <= 255:
        return np.uint8
    elif min >= 0 and max <= 65535:
        return np.uint16
    elif min >= 0 and max <= 18446744073709551615:
        return np.uint64
    else:
        return array.dtype



from pydantic import BaseModel
from pathlib import Path
import os
from ...models import Config


class InstallReqs:
    def __init__(self, logger=None):
        self.logger = logger if logger else print

    class Inputs(BaseModel):
        config: Config

    class Outputs(BaseModel):
        message: str

    def __call__(self, inputs: Inputs) -> Outputs:
        # install requirements.txt in each service folder
        for service in inputs.config:
            self.logger(f"Installing requirements for service '{service.name}'...")
            service_path = (
                inputs.config.dir
                / inputs.config.type2folder(service.type)
                / service.name
            )
            reqs_path = service_path / "requirements.txt"
            if not reqs_path.exists():
                self.logger(f"No requirements.txt found for service '{service.name}'.")
                continue
            os.system(f"pip install -r {str(reqs_path)}")
        return self.Outputs(message="done")



import numpy as np
import math
import pandas as pd

from scipy.ndimage import median_filter
from skimage.exposure import equalize_hist

from ..processing import mask_raster
from ..processing import read_raster
from ..processing import colorize_raster
from ..processing import px_count
from ..processing import save_table
from ..processing import normalised_difference
from ..processing import convert_array_to_vector

from .utils import format_name


def water_quality(image_name, date, storage, prefix="", names={}):
    """
    This function calculates the water quality of a given image.

    It calculates the following layers:
        - Normalized Difference Water Index (NDWI) - ndwi_{date}.tif, ndwi_masked_{date}.tif, ndwi_categorized_{date}.tif
        - Normalized Difference Turbidity Index (NDTI) - ndti_{date}.tif, ndti_masked_{date}.tif, ndti_categorized_{date}.tif
        - Normalized Difference Chlorophyll Index (NDCI) - ndci_{date}.tif, ndci_masked_{date}.tif, ndci_categorized_{date}.tif
        - Water Mask - water_mask_{date}.tif
        - Dissolved Organic Carbon (DOC) - DOC_{date}.tif, DOC_masked_{date}.tif, DOC_categorized_{date}.tif

    It also calculates the following analytics:
        - Water Extent (in hectares and percentage) - table_water_extent.json
        - Water Turbidity (in hectares and percentage) - table_turbidity_Ha.json, table_turbidity_percent.json
        - Water Chlorophyll (in hectares and percentage) - table_chlorophyll_Ha.json, table_chlorophyll_percent.json
        - Water DOC (in hectares and percentage) - table_DOC_Ha.json, table_DOC_percent.json

    Parameters
    ----------
    image_name : str
        The name of the image to be processed.
    date: str
        The date of the image to be processed.
    storage : Storage
        The Storage object.

    Returns
    -------
    None

    """
    default_names = {
        "ndwi": "ndwi_{date}.tif",
        "ndwi_masked": "ndwi_masked_{date}.tif",
        "ndwi_categorized": "ndwi_categorized_{date}.tif",
        "ndti": "ndti_{date}.tif",
        "ndti_masked": "ndti_masked_{date}.tif",
        "ndti_categorized": "ndti_categorized_{date}.tif",
        "ndti_categorized_rgb": "ndti_categorized_rgb_{date}.tif",
        "ndci": "ndci_{date}.tif",
        "ndci_masked": "ndci_masked_{date}.tif",
        "ndci_categorized": "ndci_categorized_{date}.tif",
        "ndci_categorized_rgb": "ndci_categorized_rgb_{date}.tif",
        "water_mask": "water_mask_{date}.tif",
        "DOC": "DOC_{date}.tif",
        "DOC_masked": "DOC_masked_{date}.tif",
        "DOC_categorized": "DOC_categorized_{date}.tif",
        "DOC_categorized_rgb": "DOC_categorized_rgb_{date}.tif",
    }
    names = {**default_names, **names}

    # read_raster image
    ds, raster = read_raster(image_name, storage, bands=[3, 4, 5, 8])

    """ LAYERS """
    # NDWI

    # calculate ndwi
    ndwi = normalised_difference(raster, [1, 4])

    # save_raster ndwi
    raster_name_ndwi = format_name(names["ndwi"], prefix, date)
    print(raster_name_ndwi)
    storage.create(ndwi, raster_name_ndwi, ds=ds)

    # NDTI
    # calculate ndti
    ndti = normalised_difference(raster, [2, 1])

    # save_raster ndti
    raster_name_ndti = format_name(names["ndti"], prefix, date)
    storage.create(ndti, raster_name_ndti, ds=ds)

    # NDCI
    # calculate ndci
    ndci = normalised_difference(raster, [3, 2])
    ndci = equalize_hist(ndci)

    # save_raster ndci
    raster_name_ndci = format_name(names["ndci"], prefix, date)
    storage.create(ndci, raster_name_ndci, ds=ds)

    # Calculate water mask and smooth it
    # threshold on 0
    water_mask = ndwi >= 0.2
    water_mask = median_filter(water_mask, size=(10, 5))

    # save_raster water_mask
    raster_name_water_mask = format_name(names["water_mask"], prefix, date)
    storage.create(water_mask, raster_name_water_mask, ds=ds)

    # DOC
    # calculate doc
    stored_images = storage.list("sentinel-2-l2a*.tif")
    doc_layers_array, doc_values_array, ds = compute_doc(stored_images, storage)

    # save_raster doc
    raster_name_doc = format_name(names["DOC"], prefix, date)
    storage.create(doc_layers_array[0], raster_name_doc, ds=ds)

    """ MASK LAYERS """
    shp = convert_array_to_vector(water_mask, storage.get_path(raster_name_water_mask))

    # mask_raster ndwi with aoi_mask
    ndwi_masked, _ = mask_raster(raster_name_ndwi, shp, storage)

    # save_raster ndwi_masked
    raster_name_ndwi_masked = format_name(names["ndwi_masked"], prefix, date)
    storage.create(ndwi_masked, raster_name_ndwi_masked, ds=ds)

    # mask_raster ndti with aoi_mask
    ndti_masked, _ = mask_raster(raster_name_ndti, shp, storage)

    # save_raster ndti_masked
    raster_name_ndti_masked = format_name(names["ndti_masked"], prefix, date)
    storage.create(ndti_masked, raster_name_ndti_masked, ds=ds)

    # mask_raster ndci with aoi_mask
    ndci_masked, _ = mask_raster(raster_name_ndci, shp, storage)

    # save_raster ndci_masked
    raster_name_ndci_masked = format_name(names["ndci_masked"], prefix, date)
    storage.create(ndci_masked, raster_name_ndci_masked, ds=ds)

    # mask_raster doc with aoi_mask
    doc_masked, _ = mask_raster(raster_name_doc, shp, storage)

    # save_raster doc_masked
    raster_name_doc_masked = format_name(names["DOC_masked"], prefix, date)
    storage.create(doc_masked, raster_name_doc_masked, ds=ds)

    """ ANALYTICS """
    # Water Extent
    extent_has = np.divide(water_mask.sum().max(), 100)
    update_extent_table(extent_has, date, storage)

    # Water Turbidity
    ndti_categories = np.digitize(ndti_masked, [-np.inf, -0.2, 0.4, np.inf])
    raster_name_ndti_categorized = format_name(names["ndti_categorized"], prefix, date)
    storage.create(ndti_categories, raster_name_ndti_categorized, ds=ds)

    ndti_colorized = colorize_raster(ndti_categories, colors=["green", "yellow", "red"])
    raster_name_ndti_colorized = format_name(
        names["ndti_categorized_rgb"], prefix, date
    )
    storage.create(ndti_colorized, raster_name_ndti_colorized, ds=ds)

    ndti_px_counted = px_count(ndti_categories, [1, 2, 3])
    ndti_has = np.divide(
        ndti_px_counted,
        100,
        out=np.zeros_like(ndti_px_counted, dtype=np.float64),
        where=100 != 0,
    )
    # save_table turbidity hectareas
    turbidity_table_name = "table_turbidity_Ha.json"
    turbidity_columns = ["Good [Has]", "Careful [Has]", "Bad [Has]", "Total [Has]"]
    save_table(
        data=ndti_has,
        columns=turbidity_columns,
        table_name=turbidity_table_name,
        date=date,
        storage=storage,
    )
    ndti_percent = [
        np.divide(ndti_has[i], ndti_has[-1]) * 100 for i in range(len(ndti_has) - 1)
    ]
    # save_table turbidity percent
    turbidity_percent_table_name = "table_turbidity_percent.json"
    turbidity_percent_columns = ["Good [%]", "Careful [%]", "Bad [%]"]
    save_table(
        data=ndti_percent,
        columns=turbidity_percent_columns,
        table_name=turbidity_percent_table_name,
        date=date,
        storage=storage,
    )

    # Water Chlorophyll
    ndci_categories = np.digitize(ndci_masked, [-np.inf, -0.3, 0.5, np.inf])
    raster_name_ndci_categorized = format_name(names["ndci_categorized"], prefix, date)
    storage.create(ndci_categories, raster_name_ndci_categorized, ds=ds)

    ndci_colorized = colorize_raster(ndci_categories, colors=["green", "yellow", "red"])
    raster_name_ndci_colorized = format_name(
        names["ndci_categorized_rgb"], prefix, date
    )
    storage.create(ndci_colorized, raster_name_ndci_colorized, ds=ds)

    ndci_px_counted = px_count(ndci_categories, [1, 2, 3])
    ndci_has = np.divide(
        ndci_px_counted,
        100,
        out=np.zeros_like(ndci_px_counted, dtype=np.float64),
        where=100 != 0,
    )
    # save_table chlorophyll hectareas
    chlorophyll_table_name = "table_chlorophyll_Ha.json"
    chlorophyll_columns = ["Good [Has]", "Careful [Has]", "Bad [Has]", "Total [Has]"]
    save_table(
        data=ndci_has,
        columns=chlorophyll_columns,
        table_name=chlorophyll_table_name,
        date=date,
        storage=storage,
    )
    ndci_percent = [
        np.divide(ndci_has[i], ndci_has[-1]) * 100 for i in range(len(ndci_has) - 1)
    ]
    # save_table chlorophyll percent
    chlorophyll_percent_table_name = "table_chlorophyll_percent.json"
    chlorophyll_percent_columns = ["Good [%]", "Careful [%]", "Bad [%]"]
    save_table(
        data=ndci_percent,
        columns=chlorophyll_percent_columns,
        table_name=chlorophyll_percent_table_name,
        date=date,
        storage=storage,
    )

    # DOC
    # Compute the mean of the doc values
    mean_mean_doc = np.mean([dict["mean_doc"] for dict in doc_values_array])
    mean_max_doc = np.mean([dict["max_doc"] for dict in doc_values_array])
    mean_min_doc = np.mean([dict["min_doc"] for dict in doc_values_array])
    mean_std_doc = np.mean([dict["std_doc"] for dict in doc_values_array])

    # Categorize the doc values
    a = 1
    min_to_N0 = -np.inf
    N0_to_N1 = mean_mean_doc + 2 * mean_std_doc / a
    N1_to_N3 = mean_mean_doc + 4 * mean_std_doc / a
    N3_to_max = np.inf

    doc_categories = np.digitize(doc_masked, [min_to_N0, N0_to_N1, N1_to_N3, N3_to_max])
    raster_name_doc_categorized = format_name(names["DOC_categorized"], prefix, date)
    storage.create(doc_categories, raster_name_doc_categorized, ds=ds)

    doc_colorized = colorize_raster(doc_categories, colors=["green", "yellow", "red"])
    raster_name_doc_colorized = format_name(names["DOC_categorized_rgb"], prefix, date)
    storage.create(doc_colorized, raster_name_doc_colorized, ds=ds)

    doc_px_counted = px_count(doc_categories, [1, 2, 3])
    doc_has = np.divide(
        doc_px_counted,
        100,
        out=np.zeros_like(doc_px_counted, dtype=np.float64),
        where=100 != 0,
    )
    # save_table DOC hectareas
    doc_table_name = "table_DOC_Ha.json"
    doc_columns = ["Good [Has]", "Careful [Has]", "Bad [Has]", "Total [Has]"]
    save_table(
        data=doc_has,
        columns=doc_columns,
        table_name=doc_table_name,
        date=date,
        storage=storage,
    )
    doc_percent = [
        np.divide(doc_has[i], doc_has[-1]) * 100 for i in range(len(doc_has) - 1)
    ]
    # save_table DOC percent
    doc_percent_table_name = "table_DOC_percent.json"
    doc_percent_columns = ["Good [%]", "Careful [%]", "Bad [%]"]
    print(doc_percent)
    save_table(
        data=doc_percent,
        columns=doc_percent_columns,
        table_name=doc_percent_table_name,
        date=date,
        storage=storage,
    )


""" 
Methods for Layers: DOC
"""


def compute_doc(stored_images, storage):
    doc_layers_array = []
    doc_values_array = []
    for path in stored_images:
        image_name = path.split("/")[-1]
        # read raster
        ds, raster = read_raster(image_name, storage, bands=[3, 4, 5, 8])

        # calculate doc for each image
        # calculate doc
        bands = [1, 2]
        bands = np.array(bands) - 1
        # Separate the bands
        band1 = raster[bands[0], :, :]
        band2 = raster[bands[1], :, :]
        # convert the bands to floats
        band1 = band1.astype(float)
        band2 = band2.astype(float)
        doc = 432 * pow(math.e, -2.24 * band1 / band2)
        mean_doc = np.nanmean(doc)
        max_doc = np.nanmax(doc)
        min_doc = np.nanmin(doc)
        std_doc = np.nanstd(doc)
        doc_values_dict = {
            "mean_doc": mean_doc,
            "max_doc": max_doc,
            "min_doc": min_doc,
            "std_doc": std_doc,
        }
        doc_values_array.append(doc_values_dict)
        doc_layers_array.append(doc)
    return doc_layers_array, doc_values_array, ds


"""
Methods for Analytics: Water Extent
"""


def update_df(df):
    df["Total [Has]"] = df["Water [Has]"].max()
    df["Not Water [Has]"] = df["Total [Has]"] - df["Water [Has]"]
    df["Percentage [%]"] = df["Water [Has]"] / df["Total [Has]"] * 100
    return df


def update_extent_table(extent_has, date, storage):
    if f"table_water_extent.json" not in storage.list():
        df = pd.DataFrame(
            {
                "Water [Has]": extent_has,
                "Not Water [Has]": 0,
                "Total [Has]": extent_has,
                "Percentage [%]": 100,
            },
            index=[date],
        )
        df = update_df(df)
    else:
        df = storage.read(f"table_water_extent.json")
        if date in df.index:
            df.loc[date, "Water [Has]"] = extent_has
            update_df(df)
            return df
        if isinstance(df.index, pd.DatetimeIndex):
            # Handle pd.DateTimeIndex, converting index to string
            df.index = df.index.strftime("%Y-%m-%d")
        new_row = pd.DataFrame(
            {
                "Water [Has]": extent_has,
                "Not Water [Has]": 0,
                "Total [Has]": 0,
                "Percentage [%]": 0,
            },
            index=[date],
        )
        df = pd.concat([new_row, df.loc[:]])
        df = update_df(df)
    storage.create(df, f"table_water_extent.json")
    return df



"""
AWSS2L2ADownloader class to download and load Sentinel-2 L2A data from a STAC API of the AWS Registry of Open Data.
"""

from .AWSDownloader import AWSDownloader

from typing import Optional, List


S2L2A_BANDS = [
    "coastal",
    "blue",
    "green",
    "red",
    "rededge1",
    "rededge2",
    "rededge3",
    "nir",
    "nir08",
    "nir09",
    "swir16",
    "swir22",
]


class AWSS2L2ADownloader(AWSDownloader):
    def __init__(
        self,
        aoi,
        datetime,
        bands: Optional[List[str]] = S2L2A_BANDS,
        resolution: Optional[float] = 0.00009009,
        crs: Optional[str] = "EPSG:4326",
        cloud_cover: Optional[float] = 100,
        vegetation_percentage: Optional[float] = 100,
        dark_features_percentage: Optional[float] = 100,
        cloud_shadow_percentage: Optional[float] = 100,
        water_percentage: Optional[float] = 100,
        not_vegetated_percentage: Optional[float] = 100,
        snow_ice_percentage: Optional[float] = 100,
        nodata_pixel_percentage: Optional[float] = 100,
    ):
        # TODO add filters
        # See filters in https://earth-search.aws.element84.com/v1/collections/sentinel-2-l2a/items
        query = {
            "eo:cloud_cover": {"lt": cloud_cover},
            "s2:vegetation_percentage": {"lt": vegetation_percentage},
            "s2:dark_features_percentage": {"lt": dark_features_percentage},
            "s2:cloud_shadow_percentage": {"lt": cloud_shadow_percentage},
            "s2:water_percentage": {"lt": water_percentage},
            "s2:not_vegetated_percentage": {"lt": not_vegetated_percentage},
            "s2:snow_ice_percentage": {"lt": snow_ice_percentage},
            "s2:nodata_pixel_percentage": {"lt": nodata_pixel_percentage},
        }
        super().__init__(aoi, datetime, query)
        self.collection = "sentinel-2-l2a"
        self.bands = bands
        self.resolution = resolution
        self.crs = crs



"""
Module to download satellite imagery data from STAC API.
"""

from .aws import AWSS2L2ADownloader
from .aws import AWSDEM30Downloader
from .aws import AWSDEM90Downloader
from .pc import PCS1GRDDownloader



from .AWSDEMDownloader import AWSDEMDownloader

from typing import Optional


class AWSDEM90Downloader(AWSDEMDownloader):
    def __init__(
        self,
        aoi,
        datetime=None,
        resolution: Optional[float] = 0.0009,
        crs: Optional[str] = "EPSG:4326",
    ):
        super().__init__(aoi, datetime)
        self.collection = "cop-dem-glo-90"
        self.datetime = datetime
        self.resolution = resolution
        self.crs = crs



"""
AWSDownloader class for downloading data from the AWS STAC API.
"""

from ..STACDownloader import STACDownloader
from typing import Optional


class AWSDownloader(STACDownloader):
    def __init__(self, aoi, datetime, query: Optional[dict] = None):
        self.url = "https://earth-search.aws.element84.com/v1"
        self.modifier = None
        super().__init__(aoi, datetime, query)



import pandas as pd
import json
import io
from minio import Minio
import fnmatch
from io import BytesIO
import numpy as np
import shutil
import os

from .BaseStorage import BaseStorage
from .decorators import with_rio, with_geopandas, with_rioxarray


# Ver funciones en CloudStorage si las de aquí no acaban de funcionar...


class S3Storage(BaseStorage):
    def __init__(self, url, access, secret, bucket, region=None):
        super().__init__()
        self.url = url
        self.access = access
        self.secret = secret
        self.region = region
        self.bucket = bucket

        if self.access and self.secret:
            # Create a client
            self.client = Minio(
                endpoint=self.url,
                access_key=self.access,
                secret_key=self.secret,
                secure=True if self.region else False,
                region=self.region,
            )  # because no certificate is used in the containerised version of minio
            if not self.client.bucket_exists(self.bucket):
                # Make a bucket with the credentials and the bucket_name given
                self.client.make_bucket(self.bucket)
                print(f"Bucket '{self.bucket}' created")
            else:
                print(f"'{self.bucket}' bucket ready to use")
        else:
            # TODO: create bucket in our minio server (we will need our credentials for that, do it with API request?
            print("Missing credentials")
            # Habría que preguntar si se quiere crear el bucket en nuestro cloud o decirles que introduzcan sus creds

    def get_path(self, name):
        return self.get_url(name)

    def get_url(self, name):
        return self.client.presigned_get_object(self.bucket, name)

    def exists(self, name):
        try:
            self.client.stat_object(self.bucket, name)
            return True
        except:
            return False

    def list(self, pattern="*"):
        return fnmatch.filter(
            [
                obj.object_name
                for obj in self.client.list_objects(self.bucket, recursive=True)
            ],
            pattern,
        )

    def create_from_path(self, data, name):
        if data.endswith(".json"):
            content_type = "application/json"
            self.client.fput_object(self.bucket, name, data, content_type=content_type)
            return self.get_url(name)
        if data.endswith(".tiff") or data.endswith(".tif"):
            content_type = "image/tiff"
            self.client.fput_object(self.bucket, name, data, content_type=content_type)
            return self.get_url(name)
        if data.endswith(".geojson"):
            content_type = "application/geojson"
            self.client.fput_object(self.bucket, name, data, content_type=content_type)
            return self.get_url(name)
        self.client.fput_object(self.bucket, name, data)
        return self.get_url(name)

    def create_from_image(self, data, name):
        image_bytes = BytesIO()
        # get format from file extension
        format = name.split(".")[-1]
        data.save(image_bytes, format=format)
        image_bytes.seek(0)
        self.client.put_object(
            self.bucket, name, image_bytes, len(image_bytes.getvalue())
        )
        return self.get_url(name)

    @with_rio
    def create_from_rasterio(self, rio, x, name, ds, window=None):
        transform = ds.transform if window is None else ds.window_transform(window)
        with rio.io.MemoryFile() as memfile:
            with memfile.open(
                driver="GTiff",
                count=1 if x.ndim < 3 else x.shape[0],
                height=x.shape[0] if x.ndim < 3 else x.shape[1],
                width=x.shape[1] if x.ndim < 3 else x.shape[2],
                dtype=np.uint8 if x.dtype == "bool" else x.dtype,
                crs=ds.crs,
                transform=transform,
                # nbits=1 if x.dtype == 'bool' else
            ) as dest_ds:
                bands = 1 if x.ndim < 3 else [i + 1 for i in range(x.shape[0])]
                dest_ds.write(x, bands)
            self.client.put_object(
                self.bucket, name, memfile, length=-1, part_size=10 * 1024 * 1024
            )
        return self.get_url(name)

    @with_rioxarray
    def create_from_xarray(self, rxr, data, name):
        # TODO don't save to disk
        content_type = "image/tiff"
        tmp_path = f"/tmp/{name}"
        self.create_dir(tmp_path)
        data.rio.to_raster(tmp_path)
        self.client.fput_object(
            self.bucket,
            name,
            tmp_path,
            content_type=content_type,
        )
        shutil.rmtree(os.path.dirname(tmp_path))
        return self.get_url(name)

    def create_from_array(self, data, name):
        array_bytes = BytesIO()
        np.save(array_bytes, data)
        array_bytes.seek(0)
        self.client.put_object(
            self.bucket, name, array_bytes, len(array_bytes.getvalue())
        )
        return self.get_url(name)

    def create_from_csv(self, data, name):
        csv_bytes = BytesIO()
        data.to_csv(csv_bytes)
        csv_bytes.seek(0)
        self.client.put_object(self.bucket, name, csv_bytes, len(csv_bytes.getvalue()))
        return self.get_url(name)

    def create_from_json(self, data, name):
        json_bytes = BytesIO()
        data.to_json(json_bytes)
        json_bytes.seek(0)
        self.client.put_object(
            self.bucket, name, json_bytes, len(json_bytes.getvalue())
        )
        return self.get_url(name)

    def create_from_dict(self, data, name):
        if name.endswith(".json"):
            content_type = "application/json"
            content = json.dumps(data, ensure_ascii=False).encode("utf8")
            self.client.put_object(
                self.bucket,
                name,
                io.BytesIO(content),
                -1,
                part_size=50 * 1024 * 1024,
                content_type=content_type,
            )
            return name
        elif name.endswith(".geojson"):
            content_type = "application/geojson"
            content = json.dumps(data, ensure_ascii=False).encode("utf8")
            self.client.put_object(
                self.bucket,
                name,
                io.BytesIO(content),
                -1,
                part_size=50 * 1024 * 1024,
                content_type=content_type,
            )
            return name
        else:
            raise TypeError("Not a valid dict type extension")

    def create_from_string(self, data, name):
        content_type = "text/plain"
        prefix = "txt/"
        content = data.encode("utf8")
        self.client.put_object(
            self.bucket,
            name,
            io.BytesIO(content),
            -1,
            part_size=50 * 1024 * 1024,
            content_type=content_type,
        )
        return name

    def create_from_dataframe(self, data, name):
        if name.endswith(".csv"):
            content_type = "text/csv"
            content = data.to_csv().encode("utf8")
            self.client.put_object(
                self.bucket,
                name,
                io.BytesIO(content),
                -1,
                part_size=50 * 1024 * 1024,
                content_type=content_type,
            )
            return name
        elif name.endswith(".json") or name.endswith(".geojson"):
            type = "json" if name.endswith(".json") else "geojson"
            content_type = f"application/{type}"
            content = data.to_json().encode("utf8")
            self.client.put_object(
                self.bucket,
                name,
                io.BytesIO(content),
                -1,
                part_size=50 * 1024 * 1024,
                content_type=content_type,
            )
            return name
        else:
            raise TypeError("Not a valid dataframe type extension")

    def read_object(self, name):
        return BytesIO(self.client.get_object(self.bucket, name).read())

    def read_from_json(self, name):
        return pd.read_json(self.read_object(name))

    def read_from_array(self, name):
        return np.load(self.read_object(name))

    @with_geopandas
    def read_from_geojson(self, gpd, name):
        try:
            response = self.client.get_object(self.bucket, name)
            data = json.load(response)
        finally:
            response.close()
            response.release_conn()
        return gpd.GeoDataFrame.from_features(data)

    @with_rio
    def read_from_rasterio(self, rio, name):
        return rio.open(self.read_object(name))

    def read_from_csv(self, name):
        return pd.read_csv(self.read_object(name), index_col=0)



import json
import os


class SPAIVars:
    def __init__(self):
        current_dir = os.getcwd()
        max_tries, current_try = 3, 1
        self.vars = None
        while current_try <= max_tries:
            file_path = current_dir + "/spai.vars.json"
            if os.path.isfile(file_path):
                print(f"Loading vars from {file_path}")
                with open(current_dir + "/spai.vars.json") as f:
                    self.vars = json.load(f)
                break
            current_try += 1
            # go up one directory
            current_dir = os.path.dirname(current_dir)
        if self.vars is None:
            print("No vars file found")

    def __getitem__(self, key):
        if self.vars and key in self.vars:
            return self.vars[key]
        else:
            raise KeyError(f"{key} not found in vars")

    def __repr__(self) -> str:
        if self.vars:
            return json.dumps(self.vars, indent=2)


def parse_vars(vars):
    variables = {}
    for var in vars:
        key, value = var.split("=")
        try:
            # Attempt to parse the value as JSON
            value = json.loads(value)
        except json.JSONDecodeError:
            # If it's not valid JSON, leave it as a string
            pass
        variables[key] = value
    return variables
